(*
  Fibonacci using TAL_i

  Nadeem Abdul Hamid, February 2006
*)

Require Import List.
Require Import tactics.
Require Import tali.

Definition halt_addr := 0.
Definition fib_entry_addr := 1.
Definition fib_body_addr := 2.

Definition halt_type : ctype :=
  cpnat (fun i =>
    cbase (fun r =>
      match r with
	| r0 => tint i
	| _ => ttop
      end)).

Definition halt_ciseq : ciseq halt_type :=
  fun i =>
    ijd halt_addr (i::nil).


Definition fib_entry_type : ctype :=
  cpnat (fun n =>
  cpnat (fun i1 =>
  cpnat (fun i2 =>
    cbase (fun r =>
      match r with
	| r0 => tint n
	| r1 => tint i1
	| r2 => tint i2
	| r7 => tcode halt_type
	| _ => ttop
      end)))).

Definition fib_entry_ciseq : ciseq fib_entry_type :=
  fun n i1 i2 =>
    icons (ibgti r0 0 fib_body_addr (n::i1::i2::nil)) (
    icons (imov r0 r1) (
    icons (itappi r7 i1) (
      ijmp r7))).


Definition fib_body_type : ctype := fib_entry_type.

Definition fib_body_ciseq : ciseq fib_body_type :=
  fun n i1 i2 =>
    icons (imov r3 r1) (
    icons (imov r1 r2) (
    icons (iadd r2 r2 r3) (
    icons (isubi r0 r0 1) (
      ijd fib_entry_addr ((n-1)::i2::(i2+i1)::nil))))) %nat.


Definition ct : forall t, ciseq t -> codeval
  := fun t cis => @existS ctype (fun t => ciseq t) t cis.

Definition CH : codeheap :=
  fun a =>
    match a with
      | 0 => Some (ct halt_type halt_ciseq)
      | 1 => Some (ct fib_entry_type fib_entry_ciseq)
      | 2 => Some (ct fib_body_type fib_body_ciseq)
      | _ => None
    end.


Lemma wf_halt : (wf_codeval CH (ct halt_type halt_ciseq)).
Proof.
  unfold ct, wf_codeval. 
  unfold halt_type, halt_ciseq; repeat (constructor; auto).
  apply wf_ijd with (ct halt_type halt_ciseq) 
    (fun r => match r with r0 => tint n | _ => ttop end);
  compute; auto.
  repeat (constructor; auto).
Qed.


Lemma wf_fib_entry : (wf_codeval CH (ct fib_entry_type fib_entry_ciseq)).
Proof.
  unfold ct, wf_codeval.
  unfold fib_entry_type, fib_entry_ciseq; repeat (constructor; auto).
  apply wf_ibgti with n (ct fib_body_type fib_body_ciseq)
    (fun r => match r with r0 => tint n | r1 => tint n0 | r2 => tint n1 
		| r7 => tcode halt_type | _ => ttop end); auto.
  compute; repeat constructor; auto.
  constructor; auto.
  repeat (constructor; auto).
  eapply wf_itappi; eauto.
  compute; eauto.
  apply wf_ijmp with 
    (fun r => match r with r0 => tint n0 | _ => ttop end).
  compute; eauto.
  intro r; elim r; compute; auto.
Qed.


Lemma wf_fib_body : (wf_codeval CH (ct fib_body_type fib_body_ciseq)).
Proof.
  unfold ct, wf_codeval.
  unfold fib_body_type, fib_body_ciseq, fib_entry_type; 
    repeat (constructor; auto).
  eapply wf_iadd; [compute; eauto | compute; eauto | idtac].
  eapply wf_isubi; [compute; eauto | idtac].
  apply wf_ijd with (ct fib_entry_type fib_entry_ciseq)
    (fun r => match r with r0 => tint (n-1) | r1 => tint n1
		| r2 => tint (n1+n0) | r7 => tcode halt_type
		| _ => ttop end).
  compute; auto.
  compute; repeat (constructor; auto).
  intro r; elim r; compute; auto.
Qed.



Lemma wf_ch : wf_codeheap CH.
Proof.
  constructor; intro f.
  elim f; clear f; [ intros; inversion H |  intros f _ ].
  apply wf_halt.

  elim f; clear f; [ intros; inversion H |  intros f _ ].
  apply wf_fib_entry.

  elim f; clear f; [ intros; inversion H |  intros f _ ].
  apply wf_fib_body.

  intros; inversion H.
Qed.


Definition RF : regfile := fun r => wint 0.
Definition RFT : rftype := fun r => tint 0.

Definition IS : iseq :=
  (icons (imovf r7 halt_addr)
  (icons (imovi r0 10)
  (icons (imovi r1 1)
  (icons (imovi r2 1)
    (ijd fib_entry_addr ( 10::1::1::nil)))))).


Goal (wf_progstate (CH, RF, IS)).
Proof.
  apply wf_ps with RFT. 
  apply wf_ch.
  constructor; intro r; elim r; compute; constructor; auto.

  unfold IS.
  eapply wf_imovf; eauto.
  compute; eauto.
  repeat (constructor; auto).
  eapply wf_ijd; eauto.
  compute; eauto.

  unfold ctypeOf, projS1; eauto.
  repeat (constructor; compute; auto).
  intro r; elim r; compute; auto.
Qed.









